@@ -1,20 +1,54 @@
-# An Object stream represents a stream of objects, floats, integers, etc.
+from __future__ import annotations
+
 import ast
-from typing import Any, Awaitable, Callable, List, Union, cast
+import logging
+from typing import (
+    Any,
+    Awaitable,
+    Callable,
+    Dict,
+    Generic,
+    Iterable,
+    List,
+    Optional,
+    Type,
+    TypeVar,
+    Union,
+    cast,
+)
 
 from make_it_sync import make_sync
 
+from func_adl.util_types import unwrap_iterable
+
 from .util_ast import as_ast, function_call, parse_as_ast
 
+# Attribute that will be used to store the executor reference
+executor_attr_name = "_func_adl_executor"
 
-# class ObjectStreamException(Exception):
-#     'Exception thrown by the ObjectStream object.'
-#     def __init__(self, msg):
-#         Exception.__init__(self, msg)
+T = TypeVar("T")
+S = TypeVar("S")
 
 
-class ObjectStream:
-    r'''
+class ReturnedDataPlaceHolder:
+    """Type returned for awkward, etc.,"""
+
+    pass
+
+
+def _local_simplification(a: ast.Lambda) -> ast.Lambda:
+    """Simplify the AST by removing unnecessary statements and
+    syntatic sugar
+    """
+    from func_adl.ast.syntatic_sugar import resolve_syntatic_sugar
+
+    r = resolve_syntatic_sugar(a)
+    assert isinstance(r, ast.Lambda), "resolve_syntatic_sugar must return a lambda"
+    return r
+
+
+class ObjectStream(Generic[T]):
+    r"""
     The objects can be events, jets, electrons, or just floats, or arrays of floats.
 
     `ObjectStream` holds onto the AST that will produce this stream of objects. The chain
@@ -26,24 +60,36 @@ class ObjectStream:
     stream is an array. You can also lift this second array of `Jets` and turn it into a plain
     stream of `Jets` using the `SelectMany` method below. In that case, you'll no longer be able
     to tell the boundary between events.
-    '''
-    def __init__(self, the_ast: ast.AST):
+
+    `TypeVar` T is the item type (so this represents, if one were to think of this as loop-a-ble,
+    `Iterable[T]`).
+    """
+
+    def __init__(self, the_ast: ast.AST, item_type: Type = Any):
         r"""
         Initialize the stream with the ast that will produce this stream of objects.
         The user will almost never use this initializer.
         """
         self._q_ast = the_ast
+        self._item_type = item_type
+
+    @property
+    def item_type(self) -> Type:
+        "Returns the type of the item this is a stream of. None if not known."
+        return self._item_type
 
     @property
     def query_ast(self) -> ast.AST:
-        '''Return the query `ast` that this `ObjectStream` represents
+        """Return the query `ast` that this `ObjectStream` represents
 
         Returns:
             ast.AST: The python `ast` that is represented by this query
-        '''
+        """
         return self._q_ast
 
-    def SelectMany(self, func: Union[str, ast.Lambda, Callable]) -> 'ObjectStream':
+    def SelectMany(
+        self, func: Union[str, ast.Lambda, Callable[[T], Iterable[S]]]
+    ) -> ObjectStream[S]:
         r"""
         Given the current stream's object type is an array or other iterable, return
         the items in this objects type, one-by-one. This has the effect of flattening a
@@ -61,10 +107,17 @@ class ObjectStream:
             - The function can be a `lambda`, the name of a one-line function, a string that
               contains a lambda definition, or a python `ast` of type `ast.Lambda`.
         """
-        return ObjectStream(function_call("SelectMany",
-                                          [self._q_ast, cast(ast.AST, parse_as_ast(func))]))
+        from func_adl.type_based_replacement import remap_from_lambda
 
-    def Select(self, f: Union[str, ast.Lambda, Callable]) -> 'ObjectStream':
+        n_stream, n_ast, rtn_type = remap_from_lambda(
+            self, _local_simplification(parse_as_ast(func))
+        )
+        return ObjectStream[S](
+            function_call("SelectMany", [n_stream.query_ast, cast(ast.AST, n_ast)]),
+            unwrap_iterable(rtn_type),
+        )
+
+    def Select(self, f: Union[str, ast.Lambda, Callable[[T], S]]) -> ObjectStream[S]:
         r"""
         Apply a transformation function to each object in the stream, yielding a new type of
         object. There is a one-to-one correspondence between the input objects and output objects.
@@ -81,10 +134,15 @@ class ObjectStream:
             - The function can be a `lambda`, the name of a one-line function, a string that
               contains a lambda definition, or a python `ast` of type `ast.Lambda`.
         """
-        return ObjectStream(function_call("Select", [self._q_ast, cast(ast.AST, parse_as_ast(f))]))
+        from func_adl.type_based_replacement import remap_from_lambda
 
-    def Where(self, filter: Union[str, ast.Lambda, Callable]) -> 'ObjectStream':
-        r'''
+        n_stream, n_ast, rtn_type = remap_from_lambda(self, _local_simplification(parse_as_ast(f)))
+        return ObjectStream[S](
+            function_call("Select", [n_stream.query_ast, cast(ast.AST, n_ast)]), rtn_type
+        )
+
+    def Where(self, filter: Union[str, ast.Lambda, Callable[[T], bool]]) -> ObjectStream[T]:
+        r"""
         Filter the object stream, allowing only items for which `filter` evaluates as true through.
 
         Arguments:
@@ -98,11 +156,69 @@ class ObjectStream:
         Notes:
             - The function can be a `lambda`, the name of a one-line function, a string that
               contains a lambda definition, or a python `ast` of type `ast.Lambda`.
-        '''
-        return ObjectStream(function_call("Where",
-                                          [self._q_ast, cast(ast.AST, parse_as_ast(filter))]))
+        """
+        from func_adl.type_based_replacement import remap_from_lambda
 
-    def AsPandasDF(self, columns=[]) -> 'ObjectStream':
+        n_stream, n_ast, rtn_type = remap_from_lambda(
+            self, _local_simplification(parse_as_ast(filter))
+        )
+        if rtn_type != bool:
+            raise ValueError(f"The Where filter must return a boolean (not {rtn_type})")
+        return ObjectStream[T](
+            function_call("Where", [n_stream.query_ast, cast(ast.AST, n_ast)]), self.item_type
+        )
+
+    def MetaData(self, metadata: Dict[str, Any]) -> ObjectStream[T]:
+        """Add metadata to the current object stream. The metadata is an arbitrary set of string
+        key-value pairs. The backend must be able to properly interpret the metadata.
+
+        Returns:
+            ObjectStream: A new stream, of the same type and contents, but with metadata added.
+        """
+        return ObjectStream[T](
+            function_call("MetaData", [self._q_ast, as_ast(metadata)]), self.item_type
+        )
+
+    def QMetaData(self, metadata: Dict[str, Any]) -> ObjectStream[T]:
+        """Add query metadata to the current object stream.
+
+        - Metadata is never transmitted to any back end
+        - Metadata is per-query, not per sample.
+
+        Warnings are issued if metadata is overwriting metadata.
+
+        Args:
+            metadata (Dict[str, Any]): Metadata to be used later
+
+        Returns:
+            ObjectStream[T]: The object stream, with metadata attached
+        """
+        from .ast.meta_data import lookup_query_metadata
+
+        first = True
+        base_ast = self.query_ast
+        for k, v in metadata.items():
+            found_md = lookup_query_metadata(self, k)
+            add_md = False
+            if found_md is None:
+                add_md = True
+            elif found_md != v:
+                logging.getLogger(__name__).info(
+                    f'Overwriting metadata "{k}" from its old value of "{found_md}" to "{v}"'
+                )
+                add_md = True
+            if add_md:
+                if first:
+                    first = False
+                    base_ast = self.MetaData({}).query_ast
+                    base_ast._q_metadata = {}  # type: ignore
+                base_ast._q_metadata[k] = v  # type: ignore
+
+        return ObjectStream[T](base_ast, self.item_type)
+
+    def AsPandasDF(
+        self, columns: Union[str, List[str]] = []
+    ) -> ObjectStream[ReturnedDataPlaceHolder]:
         r"""
         Return a pandas stream that contains one item, an pandas `DataFrame`.
         This `DataFrame` will contain all the data fed to it. Only non-array datatypes are
@@ -114,11 +230,19 @@ class ObjectStream:
                         Exception will be thrown if the number of columns do not match.
 
         """
-
         # To get Pandas use the ResultPandasDF function call.
-        return ObjectStream(function_call("ResultPandasDF", [self._q_ast, as_ast(columns)]))
+        if isinstance(columns, str):
+            columns = [columns]
 
-    def AsROOTTTree(self, filename, treename, columns=[]) -> 'ObjectStream':
+        return ObjectStream[ReturnedDataPlaceHolder](
+            function_call("ResultPandasDF", [self._q_ast, as_ast(columns)])
+        )
+
+    as_pandas = AsPandasDF
+
+    def AsROOTTTree(
+        self, filename: str, treename: str, columns: Union[str, List[str]] = []
+    ) -> ObjectStream[ReturnedDataPlaceHolder]:
         r"""
         Return the sequence of items as a ROOT TTree. Each item in the ObjectStream
         will get one entry in the file. The items must be of types that the infrastructure
@@ -143,14 +267,23 @@ class ObjectStream:
             dataset.  The order of the files back is consistent for different queries on the same
             dataset.
         """
-        return ObjectStream(
-            function_call("ResultTTree",
-                          [self._q_ast, as_ast(columns), as_ast(treename), as_ast(filename)])
-            )
+        if isinstance(columns, str):
+            columns = [columns]
 
-    def AsParquetFiles(self, filename: str, columns: Union[str, List[str]] = []) -> 'ObjectStream':
-        '''Returns the sequence of items as a `parquet` file. Each item in the ObjectStream gets a separate
-        entry in the file. The times must be of types that the infrastructure can work with:
+        return ObjectStream[ReturnedDataPlaceHolder](
+            function_call(
+                "ResultTTree", [self._q_ast, as_ast(columns), as_ast(treename), as_ast(filename)]
+            )
+        )
+
+    as_ROOT_tree = AsROOTTTree
+
+    def AsParquetFiles(
+        self, filename: str, columns: Union[str, List[str]] = []
+    ) -> ObjectStream[ReturnedDataPlaceHolder]:
+        """Returns the sequence of items as a `parquet` file. Each item in the ObjectStream
+        gets a separate entry in the file. The times must be of types that the infrastructure
+        can work with:
 
             Float               A tree with a single float in each entry will be written.
             vector<float>       A tree with a list of floats in each entry will be written.
@@ -173,12 +306,20 @@ class ObjectStream:
             written by the backend - the data should be concatinated together to get a final
             result. The order of the files back is consistent for different queries on the same
             dataset.
-        '''
-        return ObjectStream(function_call("ResultParquet",
-                                          [self._q_ast, as_ast(columns), as_ast(filename)]))
+        """
+        if isinstance(columns, str):
+            columns = [columns]
 
-    def AsAwkwardArray(self, columns=[]) -> 'ObjectStream':
-        r'''
+        return ObjectStream[ReturnedDataPlaceHolder](
+            function_call("ResultParquet", [self._q_ast, as_ast(columns), as_ast(filename)])
+        )
+
+    as_parquet = AsParquetFiles
+
+    def AsAwkwardArray(
+        self, columns: Union[str, List[str]] = []
+    ) -> ObjectStream[ReturnedDataPlaceHolder]:
+        r"""
         Return a pandas stream that contains one item, an `awkward` array, or dictionary of
         `awkward` arrays. This `awkward` will contain all the data fed to it.
 
@@ -190,12 +331,20 @@ class ObjectStream:
         Returns:
 
             An `ObjectStream` with the `awkward` array data as its one and only element.
-        '''
-        return ObjectStream(function_call("ResultAwkwardArray", [self._q_ast, as_ast(columns)]))
+        """
+        if isinstance(columns, str):
+            columns = [columns]
 
-    def _get_executor(self, executor: Callable[[ast.AST], Awaitable[Any]] = None) \
-            -> Callable[[ast.AST], Awaitable[Any]]:
-        r'''
+        return ObjectStream[ReturnedDataPlaceHolder](
+            function_call("ResultAwkwardArray", [self._q_ast, as_ast(columns)])
+        )
+
+    as_awkward = AsAwkwardArray
+
+    def _get_executor(
+        self, executor: Optional[Callable[[ast.AST, Optional[str]], Awaitable[Any]]] = None
+    ) -> Callable[[ast.AST, Optional[str]], Awaitable[Any]]:
+        r"""
         Returns an executor that can be used to run this.
         Logic seperated out as it is used from several different places.
 
@@ -204,17 +353,26 @@ class ObjectStream:
 
         Returns:
             An executor that is either synchronous or a coroutine.
-        '''
+        """
         if executor is not None:
             return executor
 
-        from .event_dataset import find_ed_in_ast
-        ed = find_ed_in_ast(self._q_ast)
+        # Dig into the AST until we find a node with an executor reference attached. The AST is
+        # traversed by looking recursively into the source of each ObjectStream, which is always
+        # the first argument in the ast.Call node.
+        node = self._q_ast
+        while not hasattr(node, executor_attr_name):
+            node = node.args[0]  # type: ignore
 
-        return ed.execute_result_async
+        # Extract the executor from this reference.
+        return getattr(node, executor_attr_name)
 
-    async def value_async(self, executor: Callable[[ast.AST], Any] = None) -> Any:
-        r'''
+    async def value_async(
+        self,
+        executor: Optional[Callable[[ast.AST, Optional[str]], Any]] = None,
+        title: Optional[str] = None,
+    ) -> Any:
+        r"""
         Evaluate the ObjectStream computation graph. Tracks back to the source dataset to
         understand how to evaluate the AST. It is possible to pass in an executor to override that
         behavior (used mostly for testing).
@@ -225,6 +383,7 @@ class ObjectStream:
                             result. If None, then uses the default executor. Normally is none
                             and the default executor specified by the `EventDatasource` is called
                             instead.
+            title           Optional title to hand to the transform
 
         Returns
 
@@ -235,11 +394,13 @@ class ObjectStream:
 
             This is the non-blocking version - it will return a future which can
             be `await`ed upon until the query is done.
-        '''
+        """
         # Fetch the executor
         exe = self._get_executor(executor)
 
         # Run it
-        return await exe(self._q_ast)
+        from func_adl.ast.meta_data import remove_empty_metadata
+
+        return await exe(remove_empty_metadata(self._q_ast), title)
 
     value = make_sync(value_async)
